/* 
 * File:   rtcc.h
 * Author: Don Stanford
 *
 * Created on July 20, 2017, 3:46 PM
 */

#ifndef RTCC_H
#define	RTCC_H

#include <xc.h>
#include <stdint.h>

/*VARIABLE DECLARATION******************************************************************************/
uint8_t *pdata;
uint8_t *pflag;
uint8_t nCount;
uint8_t counter;



unsigned char seconds=0x00,minutes=0x00,hours=0x0C,day=0x00,month=0x00,year=0x00;
unsigned char sec = 0x00, min = 0x00, hr = 0x00;

char b;
uint8_t sett;
int TimeOut = 0;
int a = 0;

/*DEFINITIONS***************************************************************************************/
#define I2C_RetryMax  100

//Slave(RTCC) Address
#define     RTCC        0x50

//Register Addresses
#define     RTCCcont    0x00
#define     RTCCsec     0x02
#define     RTCCmin     0x03
#define     RTCChr      0x04
#define     RTCCyr      0x05
#define     RTCCwk_mt   0x06
#define     RTCCtm      0x07
#define     RTCCalm     0x08

// Control/Status Register (RTCCcont)
#define     timer_flag      0x01
// 50% duty factor sec if alarm enable bit = 0

#define     alarm_flag      0x02
// 50% duty factor min if alarm enable bit = 0

#define     alarm_enable    0x04    
// 0=alarm disabled addresses 0x08-0x0F free RAM
// 1=alarm enabled address 0x08 is alarm control register

#define     mask_flag       0x08
// 0 read address 0x05-06 unmasked
// 1 read date and month directly

// bits <5:4> are the function modes
#define     clock_32kHz     & 0b001111
#define     clock_50Hz      0x10
#define     event_counter   0x20
#define     test_mode       0x30

// hold last count
#define     hold_count      0x40

// stop the count
#define     stop_count      0x80



//Clock format read from hour register add 0x04 PM = 1
 #define    am_pm           0x40 

//Set the clock for 12 or 24 hour format 12 hour = 1
#define     H12_24          0x80



//Alarm control register functions (RTCCalm)

//timer function (bits <2:0>)
#define     alm_sec         0x02
#define     alm_min         0x03
#define     alm_hr          0x04
#define     alm_day         0x05

//timer interrupt on <3>
#define     tim_int         0x08

//clock alarm function (bit <5:4>)
#define     dly             0x10 //daily alarm
#define     wdy             0x20 //weekday alarm
#define     dtd             0x30 //dated alarm

//timer alarm enable (bit <6>)
#define     tmr_enl         0x40

// alarm interrupt enable (bit <7>) (only when alarm enable in control reg.)
#define     alm_int         0x80

/*ARRAYS********************************************************************************************/
uint8_t Clock[3];       /*[0]}=hr, [1]=min, [2]=sec.*/
uint8_t SetTime[3];     /*to set time on the RTCC*/
uint8_t RTime[3];       /*to read time from the RTCC*/
uint8_t Radd[8];        /*used to send the read address to RTCC*/
uint8_t PTime[2];       /*To send pump timer to RTC RAM*/
uint8_t Memory[2];

/*BIT UNIONS****************************************************************************************/
typedef unsigned char bool;
union
{
    struct 
    {
        bool H12_H24:1;     //0x01
        bool AM_PM:1;       //0x02
        bool Hr:1;          //0x04
        bool Min:1;         //0x08
        bool Sec:1;         //0x10
        bool Day:1;         //0x20
        bool Month:1;       //0x40
        bool Year:1;        //0x80
    };
    char tstatus;
}TimeStatus;



/*FUNCTIONS*****************************************************************************************/
void TimeSet(void);

void RTCCDataSend(void);

void TimeRead(void);

void TimeSetMode(void);

void BCD2BIN(void);

void BIN2BCD(void);

void HrMinSec(void);

void MemoryRead(void);

void ClockStart(void);

#ifdef	__cplusplus
extern "C" {
#endif




#ifdef	__cplusplus
}
#endif

#endif	/* RTCC_H */

